#include <linux/module.h>
#include <linux/kernel.h>

#include <linux/malloc.h>
#include <linux/sched.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/fs.h>
#include <linux/interrupt.h>
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/fcntl.h>
#include <linux/poll.h>
#include <linux/ip.h>
#include <linux/in6.h>
#include <asm/checksum.h>

#include "softlink.h"

static const char *version = "softlink.c; v0.02 (c) Florian Baumgartner (IAM, Berne)\n";

static char nt_name[NOF*16];
static struct device nt_dev[NOF];

static struct file_operations ch_fops = {
  NULL,          /* lseek */
  ch_read,       /* read  */
  ch_write,      /* write */
  NULL,          /* readdir */
  ch_poll,       /* poll */
  ch_ioctl,      /* ioctl */
  NULL,          /* mmap */
  ch_open,       /* open */
  NULL,          /* flush */
  ch_release     /* release */
};



static unsigned int ch_poll(struct file *f, poll_table *wait)
{
  struct ch_priv *cpriv=(struct ch_priv*)f->private_data;
  struct nt_priv *npriv=(struct nt_priv*)nt_dev[cpriv->minor].priv;

  unsigned int mask=0;

#ifdef DEBUG
  printk("poll");
#endif
  
  poll_wait(f, &npriv->inq, wait);
  poll_wait(f, &npriv->outq, wait);


  if(npriv->len_w){
    mask |=POLLRDNORM|POLLIN;
  }

  if(!npriv->len_r){
    mask |=POLLWRNORM|POLLOUT;
  }

  return mask;

}



int init_module(void) 
{
  int result;
  int i;

  printk(version);

#ifdef DEBUG
  printk("init_module\n");
#endif  

  memset(nt_name,0,NOF*16);

  for(i=0;i<NOF;++i) {
    sprintf(nt_name+i*16,"sol%d",i);
    nt_dev[i].name=nt_name+i*16;
    nt_dev[i].rmem_end=0;
    nt_dev[i].rmem_start=0;
    nt_dev[i].mem_end=0;
    nt_dev[i].mem_start=0;
    nt_dev[i].base_addr=0;
    nt_dev[i].irq=0;
    nt_dev[i].start=0;
    nt_dev[i].interrupt=0;
    nt_dev[i].tbusy=0;
    nt_dev[i].init=nt_init;

    if ((result = register_netdev(&nt_dev[i])) != 0)
      return result;
  }

  if((result=register_chrdev(DEVICE_MAJOR, "sol", &ch_fops))!=0)
    return result;

  return 0;
}




int cleanup_module(void) 
{
  int i;
  int result;

#ifdef DEBUG
  printk("cleanup_module\n");
#endif  

  for(i=0;i<NOF;++i) {
    kfree(nt_dev[i].priv);
    unregister_netdev(&nt_dev[i]);
  }

  
  if((result=unregister_chrdev(DEVICE_MAJOR, "sol"))!=0)
    return result;
  
  return 0;
}


static int nt_open(struct device *dev)
{
  struct nt_priv *npriv=(struct nt_priv*)dev->priv;
  
#ifdef DEBUG
  printk("nt_open\n");
#endif  

  npriv->len_r=0;
  npriv->buff_r=kmalloc(dev->mtu,GFP_KERNEL);
  npriv->len_w=0;
  npriv->buff_w=kmalloc(dev->mtu,GFP_KERNEL);

  memcpy(dev->dev_addr,"\0SOFTLINK", ETH_ALEN);
  dev->start=1;
  dev->tbusy=0;

  MOD_INC_USE_COUNT;
  return 0;
}


static int nt_stop(struct device *dev)
{
  struct nt_priv *npriv=(struct nt_priv*)dev->priv;

#ifdef DEBUG
  printk("nt_stop\n");
#endif  
  dev->start=0;
  dev->tbusy=1;

  kfree(npriv->buff_r); npriv->len_r=0;
  kfree(npriv->buff_w); npriv->len_w=0;

  MOD_DEC_USE_COUNT;
  return 0;
}




static int nt_set_config(struct device *dev, struct ifmap *map)
{
#ifdef DEBUG
  printk("nt_set_config\n");
#endif  
  return 0;
}



/*
 * network device -> char device 
 */
static int nt_tx(struct sk_buff *skb, struct device *dev)
{
  int retval=0;
  int len;
  struct nt_priv *npriv=(struct nt_priv*)dev->priv;

#ifdef DEBUG
  printk("nt_tx (%ld)\n", dev->tbusy);
#endif  

  if(dev->tbusy==1 || npriv->len_w) {
    printk("Network device timed out\n");
    dev->tbusy=0;
    npriv->len_w=0;
    mark_bh(NET_BH);
    retval=-EBUSY;
    goto tx_done;
  }
  if(skb==NULL){

#ifdef DEBUG
    printk("nt_tx skb==NULL\n");
#endif
    return 0;
  }

  dev->tbusy=1;

  len=ETH_ZLEN < skb->len ? skb->len: ETH_ZLEN;
  dev->trans_start=jiffies;

  npriv->len_w=len;
  memcpy(npriv->buff_w,skb->data,len);

  wake_up_interruptible(&npriv->inq);

 tx_done:

  dev_kfree_skb(skb);
  return retval;
}


/*
 *  ioctl
 */
static int nt_ioctl(struct device *dev, struct ifreq *ifr, int cmd) 
{ 
#ifdef DEBUG
  printk("nt_ioctl\n");
#endif  
  return 0; 
} 



struct enet_statistics *nt_stats(struct device *dev)
{
  struct nt_priv *priv= (struct nt_priv*)dev->priv;

  return &priv->stats;
}




static int nt_rebuild_header(struct sk_buff *skb)
{
  struct device *dev=skb->dev;
  struct ethhdr *eth = (struct ethhdr *)skb->data;

#ifdef DEBUG
  printk("nt_rebuild_header\n");
#endif  

  memcpy(eth->h_source, dev->dev_addr, dev->addr_len);
  memcpy(eth->h_dest, dev->dev_addr, dev->addr_len);

  // eth->h_dest[ETH_ALEN-1]   ^= 0x01;
  
  return 0;
}




static int nt_init(struct device *dev)
{
#ifdef DEBUG
  printk("nt_init\n");
#endif  

  ether_setup(dev);

  dev->open=nt_open;
  dev->stop=nt_stop;
  dev->set_config=nt_set_config;
  dev->hard_start_xmit=nt_tx;
  dev->do_ioctl=nt_ioctl;
  dev->get_stats=nt_stats;
  dev->rebuild_header=nt_rebuild_header;
  dev->flags|=IFF_NOARP;

  dev->priv=kmalloc(sizeof(struct nt_priv), GFP_KERNEL);
  if(dev->priv==NULL){
    return -ENOMEM;
  }
  memset(dev->priv, 0, sizeof(struct nt_priv));
  return 0;
}


/*
 * char device -> network device 
 */
static void nt_rx(struct device *dev, int len, unsigned char *buf){
  struct sk_buff *skb;
  struct iphdr *ih;
  char *dat;
 
#ifdef DEBUG
  printk("nt_rx\n");
#endif
  skb=dev_alloc_skb(len+sizeof(struct ethhdr)); // +sizeof(struct ethhdr));

  if(!skb) {
    printk("[softlink] low on memory !\n");
    return;
  }

  ih=(struct iphdr*)buf;

  ih->check = 0;         /* and rebuild the checksum (ip needs it) */
  ih->check = ip_fast_csum((unsigned char *)ih,ih->ihl);

  dat=skb_put(skb,len+sizeof(struct ethhdr));

  memcpy(dat, "\0SOFTL\0SOFTL\x8\0",sizeof(struct ethhdr));
  memcpy(dat+sizeof(struct ethhdr),buf,len);

  skb->dev=dev;
  skb->protocol = eth_type_trans(skb,dev);

  skb->ip_summed = CHECKSUM_UNNECESSARY;
  netif_rx(skb);

  return;
}

/*
 * a read or write to the network interface occured
 */
static void nt_interrupt(int irq, void *dev_id, struct pt_regs * regs)
{

  struct device *dev = dev_id;
  struct nt_priv *npriv;
  int status;

  if (dev == NULL) {
    printk("Interrupt got NULL as device\n");
    return;
  }
  
  dev->interrupt = 1;
  
  npriv = (struct nt_priv *)dev->priv;
  status = npriv->status;
  
#ifdef DEBUG
  printk("nt_interrupt (%u)\n", status);
#endif
    
  
  if (status & RX_INTR) {  
    npriv->status-=RX_INTR;
    npriv->stats.rx_packets++;
    nt_rx(dev, npriv->len_r, npriv->buff_r);
    npriv->len_r=0;
  }
  
  if (status & TX_INTR) {   
    npriv->status-=TX_INTR;
    npriv->stats.tx_packets++;
    dev->tbusy = 0;
    mark_bh(NET_BH);   /* Inform upper layers. */
  }
  
  dev->interrupt = 0;

  return;
}

/*
 * the char device is opened
 */
static int ch_open(struct inode *in, struct file *f)
{
  struct ch_priv *priv;
 
#ifdef DEBUG
  printk("ch_open\n");
#endif  

  f->private_data=kmalloc(sizeof(struct ch_priv),GFP_KERNEL);
  if(f->private_data==NULL)
    return -ENOMEM;
  
  priv=(struct ch_priv*)f->private_data;

  priv->minor=MINOR(in->i_rdev);
  priv->major=MAJOR(in->i_rdev);

  f->f_op=&ch_fops;
  
  MOD_INC_USE_COUNT;
  return 0;
}


/*
 * somebody reas the char device
 */
static ssize_t ch_read(struct file *f, char *d, size_t s, loff_t *l)
{
  int result;
  struct ch_priv *cpriv=(struct ch_priv*)f->private_data;
  struct nt_priv *npriv=(struct nt_priv*)nt_dev[cpriv->minor].priv;
  struct device *ndev=&nt_dev[cpriv->minor];
  
#ifdef DEBUG
  printk("ch_read\n");
#endif  

  while(npriv->len_w==0 || npriv->buff_w==0){ /* nothing to read */
    if(f->f_flags & O_NONBLOCK) 
      return -EAGAIN;

    interruptible_sleep_on(&npriv->inq);
     
    if(signal_pending(current)) return -ERESTARTSYS;
    
  }
  result=npriv->len_w-sizeof(struct ethhdr);
  memcpy(d, npriv->buff_w+sizeof(struct ethhdr),result);
  npriv->len_w=0; 

  /* tell the networkdevice the the packet has been delivered */
  npriv->status|=TX_INTR;
  nt_interrupt(0, ndev, NULL);

  return result;
}

/*
 * stop char device
 */
static int ch_release(struct inode *inode, struct file *file)
{
#ifdef DEBUG
  printk("ch_release\n");
#endif  

  kfree(file->private_data);

  MOD_DEC_USE_COUNT;
  return 0;
}

/*
 * somebody writes to the char device
 */
static ssize_t ch_write(struct file *f, const char *d, size_t s, loff_t *l)
{
  struct ch_priv *cpriv=(struct ch_priv*)f->private_data;
  struct nt_priv *npriv=(struct nt_priv*)nt_dev[cpriv->minor].priv;
  struct device *ndev=&nt_dev[cpriv->minor];

#ifdef DEBUG
  printk("ch_write\n");
#endif  
  
  if(npriv->len_r!=0)return 0;
  
  // printk("size %d\n",s);
  memcpy(npriv->buff_r, d, s);
  npriv->len_r=s;
  npriv->status|=RX_INTR;
  nt_interrupt(0, ndev, NULL);
  return s;
}



static int ch_ioctl(struct inode *in, struct file *fi, unsigned int cmd, unsigned long arg)
{
#ifdef DEBUG
  printk("ch_write\n");
#endif  

return 0;
}
  
















